==> ert-6.0.0rc0-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: ert/storage/local_storage.py <==
from __future__ import annotations

import json
import os
import sys
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Type,
    Union,
    no_type_check,
)
from uuid import UUID, uuid4

from filelock import FileLock, Timeout
from pydantic import BaseModel

from ert.config import ErtConfig
from ert.storage.local_ensemble import LocalEnsembleAccessor, LocalEnsembleReader
from ert.storage.local_experiment import LocalExperimentAccessor, LocalExperimentReader

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


if TYPE_CHECKING:
    from ert.config.parameter_config import ParameterConfig


_LOCAL_STORAGE_VERSION = 1


class _Index(BaseModel):
    version: int = _LOCAL_STORAGE_VERSION


class LocalStorageReader:
    def __init__(self, path: Union[str, os.PathLike[str]]) -> None:
        self.path = Path(path).absolute()

        self._experiments: Union[
            Dict[UUID, LocalExperimentReader], Dict[UUID, LocalExperimentAccessor]
        ]
        self._ensembles: Union[
            Dict[UUID, LocalEnsembleReader], Dict[UUID, LocalEnsembleAccessor]
        ]
        self._index: _Index

        self.refresh()

    def refresh(self) -> None:
        self._index = self._load_index()
        self._ensembles = self._load_ensembles()  # type: ignore
        self._experiments = self._load_experiments()

    def close(self) -> None:
        for ensemble in self._ensembles.values():
            ensemble.close()
        self._ensembles.clear()
        self._experiments.clear()

    def to_accessor(self) -> LocalStorageAccessor:
        raise TypeError(str(type(self)))

    def get_experiment(self, uuid: UUID) -> LocalExperimentReader:
        return self._experiments[uuid]

    def get_ensemble(self, uuid: UUID) -> LocalEnsembleReader:
        return self._ensembles[uuid]

    def get_ensemble_by_name(
        self, name: str
    ) -> Union[LocalEnsembleReader, LocalEnsembleAccessor]:
        for ens in self._ensembles.values():
            if ens.name == name:
                return ens
        raise KeyError(f"Ensemble with name '{name}' not found")

    @property
    def experiments(self) -> Generator[LocalExperimentReader, None, None]:
        yield from self._experiments.values()

    @property
    def ensembles(self) -> Generator[LocalEnsembleReader, None, None]:
        yield from self._ensembles.values()

    def _load_index(self) -> _Index:
        try:
            return _Index.parse_file(self.path / "index.json")
        except FileNotFoundError:
            return _Index()

    @no_type_check
    def _load_ensembles(self):
        try:
            ensembles = []
            for ensemble_path in (self.path / "ensembles").iterdir():
                ensemble = self._load_ensemble(ensemble_path)
                ensembles.append(ensemble)

            # Make sure that the ensembles are sorted by name in reverse. Given
            # multiple ensembles with a common name, iterating over the ensemble
            # dictionary will yield the newest ensemble first.
            ensembles = sorted(ensembles, key=lambda x: x.started_at, reverse=True)
            return {
                x.id: x
                for x in sorted(ensembles, key=lambda x: x.started_at, reverse=True)
            }
        except FileNotFoundError:
            return {}

    def _load_ensemble(self, path: Path) -> Any:
        return LocalEnsembleReader(self, path)

    def _load_experiments(self) -> Dict[UUID, LocalExperimentReader]:
        experiment_ids = {ens.experiment_id for ens in self._ensembles.values()}
        return {exp_id: self._load_experiment(exp_id) for exp_id in experiment_ids}

    def _load_experiment(self, uuid: UUID) -> LocalExperimentReader:
        return LocalExperimentReader(self, uuid, self._experiment_path(uuid))

    def _ensemble_path(self, ensemble_id: UUID) -> Path:
        return self.path / "ensembles" / str(ensemble_id)

    def _experiment_path(self, experiment_id: UUID) -> Path:
        return self.path / "experiments" / str(experiment_id)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exception: Exception,
        exception_type: Type[Exception],
        traceback: TracebackType,
    ) -> None:
        self.close()


class LocalStorageAccessor(LocalStorageReader):
    LOCK_TIMEOUT = 5

    def __init__(
        self,
        path: Union[str, os.PathLike[str]],
        *,
        ignore_migration_check: bool = False,
    ) -> None:
        self.path = Path(path)
        if not ignore_migration_check and local_storage_needs_migration(self.path):
            from ert.storage.migration.block_fs import migrate  # pylint: disable=C0415

            migrate(self.path)

        self.path.mkdir(parents=True, exist_ok=True)

        # ERT 4 checks that this file exists and if it exists tells the user
        # that their ERT storage is incompatible
        try:
            (self.path / ".fs_version").symlink_to("index.json")
        except FileExistsError:
            pass

        self._lock = FileLock(self.path / "storage.lock")
        try:
            self._lock.acquire(timeout=self.LOCK_TIMEOUT)
        except Timeout:
            raise TimeoutError(
                f"Not able to acquire lock for: {self.path}."
                " You may already be running ERT,"
                " or another user is using the same ENSPATH."
            )

        super().__init__(path)

        self._save_index()

    def close(self) -> None:
        self._save_index()
        super().close()

        if self._lock.is_locked:
            self._lock.release()
            (self.path / "storage.lock").unlink()

    def to_accessor(self) -> LocalStorageAccessor:
        return self

    def create_experiment(
        self, parameters: Optional[List[ParameterConfig]] = None
    ) -> LocalExperimentAccessor:
        exp_id = uuid4()
        path = self._experiment_path(exp_id)
        path.mkdir(parents=True, exist_ok=False)
        exp = LocalExperimentAccessor(self, exp_id, path, parameters=parameters)
        self._experiments[exp.id] = exp
        return exp

    def create_ensemble(
        self,
        experiment: Union[LocalExperimentReader, LocalExperimentAccessor, UUID],
        *,
        ensemble_size: int,
        iteration: int = 0,
        name: Optional[str] = None,
        prior_ensemble: Optional[Union[LocalEnsembleReader, UUID]] = None,
    ) -> LocalEnsembleAccessor:
        if isinstance(experiment, UUID):
            experiment_id = experiment
        else:
            experiment_id = experiment.id

        uuid = uuid4()
        path = self._ensemble_path(uuid)
        path.mkdir(parents=True, exist_ok=False)

        prior_ensemble_id: Optional[UUID] = None
        if isinstance(prior_ensemble, UUID):
            prior_ensemble_id = prior_ensemble
        elif isinstance(prior_ensemble, LocalEnsembleReader):
            prior_ensemble_id = prior_ensemble.id

        ens = LocalEnsembleAccessor.create(
            self,
            path,
            uuid,
            ensemble_size=ensemble_size,
            experiment_id=experiment_id,
            iteration=iteration,
            name=str(name),
            prior_ensemble_id=prior_ensemble_id,
        )
        self._ensembles[ens.id] = ens
        return ens

    def _save_index(self) -> None:
        if not hasattr(self, "_index"):
            return

        with open(self.path / "index.json", mode="w", encoding="utf-8") as f:
            print(self._index.json(), file=f)

    def _load_experiment(self, uuid: UUID) -> LocalExperimentAccessor:
        return LocalExperimentAccessor(self, uuid, self._experiment_path(uuid))

    def _load_ensemble(self, path: Path) -> LocalEnsembleAccessor:
        return LocalEnsembleAccessor(self, path)


def local_storage_needs_migration(path: os.PathLike[str]) -> bool:
    """
    Checks whether the path points to a LocalStorage that is in need of
    migration. Returns true if the LocalStorage is outdated OR if the path does
    not point to a LocalStorage directory (eg. it is a URL that points to a ERT
    Storage Server)
    """
    path = Path(path)

    if not path.exists():
        return False

    try:
        with open("index.json", encoding="utf-8") as f:
            version = json.load(f)["version"]
        if version == _LOCAL_STORAGE_VERSION:
            return False
        elif version < _LOCAL_STORAGE_VERSION:
            return True
        elif version > _LOCAL_STORAGE_VERSION:
            raise NotImplementedError("Incompatible ERT Local Storage")
    except KeyError as exc:
        raise NotImplementedError("Incompatible ERT Local Storage") from exc
    except FileNotFoundError:
        pass

    return _is_block_storage(path)


_migration_ert_config: Optional[ErtConfig] = None


def local_storage_set_ert_config(ert_config: Optional[ErtConfig]) -> None:
    """
    Set the ErtConfig for migration hints
    """
    global _migration_ert_config  # pylint: disable=W0603
    _migration_ert_config = ert_config


def local_storage_get_ert_config() -> ErtConfig:
    assert (
        _migration_ert_config is not None
    ), "Use 'local_storage_set_ert_config' before retrieving the config"
    return _migration_ert_config


def _is_block_storage(path: Path) -> bool:
    """Looks for ert_fstab in subdirectories"""
    for subpath in path.iterdir():
        if subpath.name.startswith("_"):
            continue

        if (subpath / "ert_fstab").exists():
            return True

    return False

==> ert-6.0.3-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: ert/storage/local_storage.py <==
from __future__ import annotations

import json
import logging
import os
import sys
from datetime import datetime
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    MutableSequence,
    Optional,
    Tuple,
    Type,
    Union,
    no_type_check,
)
from uuid import UUID, uuid4

from filelock import FileLock, Timeout
from pydantic import BaseModel, Field

from ert.config import ErtConfig
from ert.shared import __version__
from ert.storage.local_ensemble import LocalEnsembleAccessor, LocalEnsembleReader
from ert.storage.local_experiment import LocalExperimentAccessor, LocalExperimentReader

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


if TYPE_CHECKING:
    from ert.config.parameter_config import ParameterConfig


logger = logging.getLogger(__name__)

_LOCAL_STORAGE_VERSION = 2


class _Migrations(BaseModel):
    ert_version: str = __version__
    timestamp: datetime = Field(default_factory=datetime.now)
    name: str
    version_range: Tuple[int, int]


class _Index(BaseModel):
    version: int = _LOCAL_STORAGE_VERSION
    migrations: MutableSequence[_Migrations] = Field(default_factory=list)


class LocalStorageReader:
    def __init__(self, path: Union[str, os.PathLike[str]]) -> None:
        self.path = Path(path).absolute()

        self._experiments: Union[
            Dict[UUID, LocalExperimentReader], Dict[UUID, LocalExperimentAccessor]
        ]
        self._ensembles: Union[
            Dict[UUID, LocalEnsembleReader], Dict[UUID, LocalEnsembleAccessor]
        ]
        self._index: _Index

        self.refresh()

    def refresh(self) -> None:
        self._index = self._load_index()
        self._ensembles = self._load_ensembles()  # type: ignore
        self._experiments = self._load_experiments()

    def close(self) -> None:
        for ensemble in self._ensembles.values():
            ensemble.close()
        self._ensembles.clear()
        self._experiments.clear()

    def to_accessor(self) -> LocalStorageAccessor:
        raise TypeError(str(type(self)))

    def get_experiment(self, uuid: UUID) -> LocalExperimentReader:
        return self._experiments[uuid]

    def get_ensemble(self, uuid: UUID) -> LocalEnsembleReader:
        return self._ensembles[uuid]

    def get_ensemble_by_name(
        self, name: str
    ) -> Union[LocalEnsembleReader, LocalEnsembleAccessor]:
        for ens in self._ensembles.values():
            if ens.name == name:
                return ens
        raise KeyError(f"Ensemble with name '{name}' not found")

    @property
    def experiments(self) -> Generator[LocalExperimentReader, None, None]:
        yield from self._experiments.values()

    @property
    def ensembles(self) -> Generator[LocalEnsembleReader, None, None]:
        yield from self._ensembles.values()

    def _load_index(self) -> _Index:
        try:
            return _Index.parse_file(self.path / "index.json")
        except FileNotFoundError:
            return _Index()

    @no_type_check
    def _load_ensembles(self):
        try:
            ensembles = []
            for ensemble_path in (self.path / "ensembles").iterdir():
                ensemble = self._load_ensemble(ensemble_path)
                ensembles.append(ensemble)

            # Make sure that the ensembles are sorted by name in reverse. Given
            # multiple ensembles with a common name, iterating over the ensemble
            # dictionary will yield the newest ensemble first.
            ensembles = sorted(ensembles, key=lambda x: x.started_at, reverse=True)
            return {
                x.id: x
                for x in sorted(ensembles, key=lambda x: x.started_at, reverse=True)
            }
        except FileNotFoundError:
            return {}

    def _load_ensemble(self, path: Path) -> Any:
        return LocalEnsembleReader(self, path)

    def _load_experiments(self) -> Dict[UUID, LocalExperimentReader]:
        experiment_ids = {ens.experiment_id for ens in self._ensembles.values()}
        return {exp_id: self._load_experiment(exp_id) for exp_id in experiment_ids}

    def _load_experiment(self, uuid: UUID) -> LocalExperimentReader:
        return LocalExperimentReader(self, uuid, self._experiment_path(uuid))

    def _ensemble_path(self, ensemble_id: UUID) -> Path:
        return self.path / "ensembles" / str(ensemble_id)

    def _experiment_path(self, experiment_id: UUID) -> Path:
        return self.path / "experiments" / str(experiment_id)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exception: Exception,
        exception_type: Type[Exception],
        traceback: TracebackType,
    ) -> None:
        self.close()


class LocalStorageAccessor(LocalStorageReader):
    LOCK_TIMEOUT = 5

    def __init__(
        self,
        path: Union[str, os.PathLike[str]],
        *,
        ignore_migration_check: bool = False,
    ) -> None:
        self.path = Path(path)
        if not ignore_migration_check:
            try:
                version = _storage_version(self.path)
                if version == 0:
                    from ert.storage.migration import block_fs  # pylint: disable=C0415

                    block_fs.migrate(self.path)
                    self._add_migration_information(0, "block_fs")
                elif version == 1:
                    from ert.storage.migration import gen_kw  # pylint: disable=C0415

                    gen_kw.migrate(self.path)
                    self._add_migration_information(1, "gen_kw")
            except Exception as err:  # pylint: disable=broad-exception-caught
                logger.error(f"Migrating storage at {self.path} failed with {err}")

        self.path.mkdir(parents=True, exist_ok=True)

        # ERT 4 checks that this file exists and if it exists tells the user
        # that their ERT storage is incompatible
        try:
            (self.path / ".fs_version").symlink_to("index.json")
        except FileExistsError:
            pass

        self._lock = FileLock(self.path / "storage.lock")
        try:
            self._lock.acquire(timeout=self.LOCK_TIMEOUT)
        except Timeout:
            raise TimeoutError(
                f"Not able to acquire lock for: {self.path}."
                " You may already be running ERT,"
                " or another user is using the same ENSPATH."
            )

        super().__init__(path)

        self._save_index()

    def close(self) -> None:
        self._save_index()
        super().close()

        if self._lock.is_locked:
            self._lock.release()
            (self.path / "storage.lock").unlink()

    def to_accessor(self) -> LocalStorageAccessor:
        return self

    def create_experiment(
        self, parameters: Optional[List[ParameterConfig]] = None
    ) -> LocalExperimentAccessor:
        exp_id = uuid4()
        path = self._experiment_path(exp_id)
        path.mkdir(parents=True, exist_ok=False)
        exp = LocalExperimentAccessor(self, exp_id, path, parameters=parameters)
        self._experiments[exp.id] = exp
        return exp

    def create_ensemble(
        self,
        experiment: Union[LocalExperimentReader, LocalExperimentAccessor, UUID],
        *,
        ensemble_size: int,
        iteration: int = 0,
        name: Optional[str] = None,
        prior_ensemble: Optional[Union[LocalEnsembleReader, UUID]] = None,
    ) -> LocalEnsembleAccessor:
        if isinstance(experiment, UUID):
            experiment_id = experiment
        else:
            experiment_id = experiment.id

        uuid = uuid4()
        path = self._ensemble_path(uuid)
        path.mkdir(parents=True, exist_ok=False)

        prior_ensemble_id: Optional[UUID] = None
        if isinstance(prior_ensemble, UUID):
            prior_ensemble_id = prior_ensemble
        elif isinstance(prior_ensemble, LocalEnsembleReader):
            prior_ensemble_id = prior_ensemble.id

        ens = LocalEnsembleAccessor.create(
            self,
            path,
            uuid,
            ensemble_size=ensemble_size,
            experiment_id=experiment_id,
            iteration=iteration,
            name=str(name),
            prior_ensemble_id=prior_ensemble_id,
        )
        self._ensembles[ens.id] = ens
        return ens

    def _add_migration_information(self, from_version: int, name: str) -> None:
        self._index.migrations.append(
            _Migrations(
                version_range=(from_version, _LOCAL_STORAGE_VERSION),
                name=name,
            )
        )
        self._save_index()

    def _save_index(self) -> None:
        if not hasattr(self, "_index"):
            return

        with open(self.path / "index.json", mode="w", encoding="utf-8") as f:
            print(self._index.json(), file=f)

    def _load_experiment(self, uuid: UUID) -> LocalExperimentAccessor:
        return LocalExperimentAccessor(self, uuid, self._experiment_path(uuid))

    def _load_ensemble(self, path: Path) -> LocalEnsembleAccessor:
        return LocalEnsembleAccessor(self, path)


def _storage_version(path: Path) -> Optional[int]:
    if not path.exists():
        return None
    try:
        with open(path / "index.json", encoding="utf-8") as f:
            return int(json.load(f)["version"])
    except KeyError as exc:
        raise NotImplementedError("Incompatible ERT Local Storage") from exc
    except FileNotFoundError:
        if _is_block_storage(path):
            return 0
    raise ValueError("Unknown storage version")


_migration_ert_config: Optional[ErtConfig] = None


def local_storage_set_ert_config(ert_config: Optional[ErtConfig]) -> None:
    """
    Set the ErtConfig for migration hints
    """
    global _migration_ert_config  # pylint: disable=W0603
    _migration_ert_config = ert_config


def local_storage_get_ert_config() -> ErtConfig:
    assert (
        _migration_ert_config is not None
    ), "Use 'local_storage_set_ert_config' before retrieving the config"
    return _migration_ert_config


def _is_block_storage(path: Path) -> bool:
    """Looks for ert_fstab in subdirectories"""
    for subpath in path.iterdir():
        if subpath.name.startswith("_"):
            continue

        if (subpath / "ert_fstab").exists():
            return True

    return False
